# Future
from __future__ import division, print_function, unicode_literals

# Third Party
from future.utils import python_2_unicode_compatible

# Local
from .base import BaseAPIClient, BaseAPIObject


@python_2_unicode_compatible
class User(BaseAPIObject):
    """A documentcloud user"""

    api_path = "users"
    writable_fields = ["organization"]

    def __str__(self):
        return self.username


class UserClient(BaseAPIClient):
    """Client for interacting with users"""

    api_path = "users"
    resource = User



# Future
from __future__ import division, print_function, unicode_literals

# Third Party
from future.utils import python_2_unicode_compatible

# Local
from .base import BaseAPIClient, BaseAPIObject


@python_2_unicode_compatible
class Organization(BaseAPIObject):
    """A documentcloud organization"""

    api_path = "organizations"
    writable_fields = []

    def __str__(self):
        return self.name


class OrganizationClient(BaseAPIClient):
    """Client for interacting with organizations"""

    api_path = "organizations"
    resource = Organization



# Future
from __future__ import division, print_function, unicode_literals

# Standard Library
from builtins import str
from copy import copy

# Third Party
from dateutil.parser import parse as dateparser
from future.utils import python_2_unicode_compatible

# Local
from .exceptions import DuplicateObjectError
from .toolbox import get_id, merge_dicts

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence


@python_2_unicode_compatible
class APIResults(Sequence):
    """Class for encapsulating paginated list results from the API"""

    def __init__(
        self, resource, client, response, extra=None, next_=None, previous=None
    ):
        if extra is None:
            extra = {}

        self.resource = resource
        self.client = client
        json = response.json()

        self.count = json["count"]
        self.next_url = json["next"]
        self.previous_url = json["previous"]
        self._next = next_
        self._previous = previous
        self.results = [
            resource(client, merge_dicts(r, extra)) for r in json["results"]
        ]

    def __repr__(self):
        return "<APIResults: {!r}".format(self.results)  # pragma: no cover

    def __str__(self):
        return "[{}]".format(", ".join(str(r) for r in self.results))

    def __getitem__(self, key):
        # pylint: disable=unsubscriptable-object
        if key >= len(self):
            raise IndexError

        length = len(self.results)
        if key < length:
            return self.results[key]
        else:
            return self.next[key - length]

    def __len__(self):
        return self.count

    def __iter__(self):
        for result in self.results:
            yield result
        if self.next_url:
            # pylint: disable=not-an-iterable
            for result in self.next:
                yield result
        else:
            return

    def _fetch(self, url, next_=None, previous=None):
        if url:
            response = self.client.get(url, full_url=True)
            return APIResults(
                self.resource, self.client, response, next_=next_, previous=previous
            )
        else:
            return None

    @property
    def next(self):
        if self._next is None:
            self._next = self._fetch(self.next_url, previous=self)
        return self._next

    @property
    def previous(self):
        if self._previous is None:
            self._previous = self._fetch(self.previous_url, next_=self)
        return self._previous


class BaseAPIClient(object):
    """Base client for all API resources"""

    # subclasses should set these
    api_path = None
    resource = None

    def __init__(self, client):
        self.client = client

    def get(self, id_, expand=None):
        """Get a resource by its ID"""
        if expand is not None:
            params = {"expand": ",".join(expand)}
        else:
            params = {}
        response = self.client.get(
            "{}/{}/".format(self.api_path, get_id(id_)), params=params
        )
        # pylint: disable=not-callable
        return self.resource(self.client, response.json())

    def delete(self, id_):
        """Deletes a resource"""
        self.client.delete("{}/{}/".format(self.api_path, get_id(id_)))

    def all(self, **params):
        return self.list(**params)

    def list(self, **params):
        response = self.client.get(self.api_path + "/", params=params)
        return APIResults(self.resource, self.client, response)


class ChildAPIClient(BaseAPIClient):
    """Base client for sub resources"""

    def __init__(self, client, parent):
        super(ChildAPIClient, self).__init__(client)
        self.parent = parent

    def list(self, **params):
        response = self.client.get(self.api_path + "/", params=params)
        parent_name = self.parent.__class__.__name__.lower()
        return APIResults(
            self.resource, self.client, response, {parent_name: self.parent}
        )

    # try to emulate old behavior by making it act as the list of returned resources
    def __iter__(self):
        return iter(self.list())

    def __len__(self):
        return len(self.list())

    def __getitem__(self, key):
        return self.list()[key]


class BaseAPIObject(object):
    """Base object for all API resources"""

    date_fields = []

    def __init__(self, client, dict_):
        self.__dict__ = dict_
        self._client = client
        for field in self.date_fields:
            setattr(self, field, dateparser(getattr(self, field)))

    def __repr__(self):
        return "<{}: {} - {}>".format(
            self.__class__.__name__, self.id, self
        )  # pragma: no cover

    def __eq__(self, obj):
        return isinstance(obj, type(self)) and self.id == obj.id

    def put(self):
        """Alias for save"""
        return self.save()

    def save(self):
        data = {f: getattr(self, f) for f in self.writable_fields if hasattr(self, f)}
        self._client.put("{}/{}/".format(self.api_path, self.id), json=data)

    def delete(self):
        self._client.delete("{}/{}/".format(self.api_path, self.id))


@python_2_unicode_compatible
class APISet(list):
    def __init__(self, iterable, resource):
        super(APISet, self).__init__(iterable)
        self.resource = resource
        if not all(isinstance(obj, self.resource) for obj in self):
            raise TypeError(
                "Only {} can be added to this list".format(
                    self.resource.__class__.__name__
                )
            )
        ids = [obj.id for obj in self]
        for id_ in ids:
            if ids.count(id_) > 1:
                raise DuplicateObjectError(
                    "Object with ID {} appears in the list more than once".format(id_)
                )

    def append(self, obj):
        if not isinstance(obj, self.resource):
            raise TypeError(
                "Only {} can be added to this list".format(
                    self.resource.__class__.__name__
                )
            )
        if obj.id in [i.id for i in self]:
            raise DuplicateObjectError(
                "Object with ID {} appears in the list more than once".format(obj.id)
            )
        super(APISet, self).append(copy(obj))

    def add(self, obj):
        if not isinstance(obj, self.resource):
            raise TypeError(
                "Only {} can be added to this list".format(
                    self.resource.__class__.__name__
                )
            )
        # skip duplicates silently
        if obj.id not in [i.id for i in self]:
            super(APISet, self).append(copy(obj))

    def extend(self, list_):
        if not all(isinstance(obj, self.resource) for obj in list_):
            raise TypeError(
                "Only {} can be added to this list".format(
                    self.resource.__class__.__name__
                )
            )
        ids = [obj.id for obj in self + list_]
        for id_ in ids:
            if ids.count(id_) > 1:
                raise DuplicateObjectError(
                    "Object with ID {} appears in the list more than once".format(id)
                )
        super(APISet, self).extend(copy(obj) for obj in list_)



# Future
from __future__ import division, print_function, unicode_literals

PER_PAGE_MAX = 100
BULK_LIMIT = 25
BASE_URI = "https://api.beta.documentcloud.org/api/"
AUTH_URI = "https://accounts.muckrock.com/api/"
TIMEOUT = 10
RATE_LIMIT = 10
RATE_PERIOD = 1



# Future
from __future__ import division, print_function, unicode_literals

# Third Party
from future.utils import python_2_unicode_compatible
from listcrunch.listcrunch import uncrunch

# Local
from .base import BaseAPIObject, ChildAPIClient
from .toolbox import merge_dicts


@python_2_unicode_compatible
class Annotation(BaseAPIObject):
    """A note on a document"""

    writable_fields = [
        "access",
        "content",
        "page_number",
        "title",
        "x1",
        "x2",
        "y1",
        "y2",
    ]

    def __str__(self):
        return self.title

    @property
    def api_path(self):
        return "documents/{}/notes".format(self.document.id)

    @property
    def location(self):
        page_spec = uncrunch(self.document.page_spec)
        width, height = page_spec[self.page_number].split("x")
        width, height = float(width), float(height)
        # normalize to a width of 700
        height = (700 / width) * height
        width = 700
        return Location(
            int(self.y1 * height),
            int(self.x2 * width),
            int(self.y2 * height),
            int(self.x1 * width),
        )

    @property
    def page(self):
        return self.page_number

    @property
    def description(self):
        return self.content


class Location(object):
    def __init__(self, top, right, bottom, left):
        self.top = top
        self.right = right
        self.bottom = bottom
        self.left = left


class AnnotationClient(ChildAPIClient):
    """Client for interacting with Sections"""

    resource = Annotation

    @property
    def api_path(self):
        return "documents/{}/notes".format(self.parent.id)

    def create(
        self,
        title,
        page_number,
        content="",
        access="private",
        x1=None,
        y1=None,
        x2=None,
        y2=None,
    ):
        coords = [x1, y2, x2, y2]
        if not (all(c is None for c in coords) or all(c is not None for c in coords)):
            raise ValueError(
                "x1, y2, x2, y2 must either all be None or all be not None"
            )
        if coords[0] is not None and not all(0 <= c <= 1.0 for c in coords):
            raise ValueError("x1, y2, x2, y2 must all be between 0.0 and 1.0")

        data = {
            "title": title,
            "page_number": page_number,
            "content": content,
            "access": access,
            "x1": x1,
            "y1": y1,
            "x2": x2,
            "y2": y2,
        }
        response = self.client.post(self.api_path + "/", json=data)
        return Annotation(
            self.client, merge_dicts(response.json(), {"document": self.parent})
        )



# Future
from __future__ import division, print_function, unicode_literals

# Third Party
from future.utils import python_2_unicode_compatible

# Local
from .base import APISet, BaseAPIClient, BaseAPIObject
from .constants import BULK_LIMIT, PER_PAGE_MAX
from .documents import Document
from .exceptions import DoesNotExistError, MultipleObjectsReturnedError
from .toolbox import get_id, grouper


@python_2_unicode_compatible
class Project(BaseAPIObject):
    """A documentcloud project"""

    api_path = "projects"
    writable_fields = ["description", "private", "title"]

    def __init__(self, *args, **kwargs):
        per_page = kwargs.pop("per_page", PER_PAGE_MAX)
        super(Project, self).__init__(*args, **kwargs)
        self._document_list = None
        self._per_page = per_page

    def __str__(self):
        return self.title

    def save(self):
        """Add the documents to the project as well"""
        super(Project, self).save()
        if self._document_list:
            data = [{"document": d} for d in self.document_ids]
            for data_group in grouper(data, BULK_LIMIT):
                # Grouper will put None's on the end of the last group
                data_group = [d for d in data_group if d is not None]
                self._client.put(
                    "{}/{}/documents/".format(self.api_path, self.id), json=data_group
                )

    @property
    def document_list(self):
        if self._document_list is None:
            response = self._client.get(
                "{}/{}/documents/".format(self.api_path, get_id(self.id)),
                params={"per_page": self._per_page, "expand": ["document"]},
            )
            json = response.json()
            next_url = json["next"]
            results = json["results"]
            while next_url:
                response = self._client.get(next_url, full_url=True)
                json = response.json()
                next_url = json["next"]
                results.extend(json["results"])
            self._document_list = APISet(
                (Document(self._client, r["document"]) for r in results), Document
            )
        return self._document_list

    @document_list.setter
    def document_list(self, value):
        if value is None:
            self._document_list = APISet([], Document)
        elif isinstance(value, list):
            self._document_list = APISet(value, Document)
        else:
            raise TypeError("document_list must be set to a list or None")

    @property
    def documents(self):
        return self.document_list

    @documents.setter
    def documents(self, value):
        self.document_list = value

    @property
    def document_ids(self):
        return [d.id for d in self.document_list]

    def get_document(self, doc_id):
        response = self._client.get(
            "{}/{}/documents/{}".format(self.api_path, get_id(self.id), doc_id),
            params={"expand": ["document"]},
        )
        return Document(self._client, response.json()["document"])


class ProjectClient(BaseAPIClient):
    """Client for interacting with projects"""

    api_path = "projects"
    resource = Project

    # all is overriden to filter by the current user for backward compatibility
    def all(self, **params):
        return self.list(user=self.client.user_id, **params)

    def get(self, id=None, title=None):
        # pylint:disable=redefined-builtin, arguments-differ
        # pylint disables are necessary for backward compatibility
        if id is not None and title is not None:
            raise ValueError(
                "You can only retrieve a Project by id or title, not by both"
            )
        elif id is None and title is None:
            raise ValueError("You must provide an id or a title to make a request.")

        if id is not None:
            return self.get_by_id(id)
        else:
            return self.get_by_title(title)

    def get_by_id(self, id_):
        return super(ProjectClient, self).get(id_)

    def get_by_title(self, title):
        response = self.client.get(
            self.api_path + "/", params={"title": title, "user": self.client.user_id}
        )
        json = response.json()
        if json["count"] == 0:
            raise DoesNotExistError(response=response)
        elif json["count"] > 1:
            raise MultipleObjectsReturnedError(response=response)

        return self.resource(self.client, json["results"][0])

    def create(self, title, description="", private=True, document_ids=None):
        data = {"title": title, "description": description, "private": private}
        response = self.client.post(self.api_path + "/", json=data)
        project = Project(self.client, response.json())
        if document_ids:
            data = [{"document": d} for d in document_ids]
            response = self.client.put(
                "{}/{}/documents/".format(self.api_path, project.id), json=data
            )
        return project

    def get_or_create_by_title(self, title):
        try:
            project = self.get(title=title)
            created = False
        except DoesNotExistError:
            project = self.create(title=title)
            created = True
        return project, created



# Future
from __future__ import division, print_function, unicode_literals

# Third Party
from future.utils import python_2_unicode_compatible

# Local
from .base import BaseAPIObject, ChildAPIClient
from .toolbox import merge_dicts


@python_2_unicode_compatible
class Section(BaseAPIObject):
    """A section of a document"""

    writable_fields = ["page_number", "title"]

    def __str__(self):
        return "{} - p{}".format(self.title, self.page)

    @property
    def api_path(self):
        return "documents/{}/sections".format(self.document.id)

    @property
    def page(self):
        return self.page_number


class SectionClient(ChildAPIClient):
    """Client for interacting with Sections"""

    resource = Section

    @property
    def api_path(self):
        return "documents/{}/sections".format(self.parent.id)

    def create(self, title, page_number):
        data = {"title": title, "page_number": page_number}
        response = self.client.post(self.api_path + "/", json=data)
        return Section(
            self.client, merge_dicts(response.json(), {"document": self.parent})
        )



"""
A few toys the API will use.
"""
# Future
from __future__ import division, print_function, unicode_literals

# Third Party
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

try:
    from urllib.parse import urlparse
    from itertools import zip_longest
except ImportError:
    from urlparse import urlparse
    from itertools import izip_longest as zip_longest


def requests_retry_session(
    retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504), session=None
):
    """Automatic retries for HTTP requests
    See: https://www.peterbe.com/plog/best-practice-with-retries-with-requests
    """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def get_id(id_):
    """Allow specifying old or new style IDs and convert old style to new style IDs

    Old style: 123-the-slug
    New style: 123 or the-slug-123
    """

    if isinstance(id_, int):
        return id_
    elif "-" in id_:
        front = id_.split("-", 1)[0]
        if front.isdigit():
            return front
        back = id_.rsplit("-", 1)[-1]
        if back.isdigit():
            return back

    return id_


def is_url(url):
    """Is `url` a valid url?"""
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except (ValueError, AttributeError):
        return False


def grouper(iterable, num, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * num
    return zip_longest(*args, fillvalue=fillvalue)


def merge_dicts(*dicts):
    merged = {}
    for dict_ in dicts:
        merged.update(dict_)
    return merged



# Local
from .annotations import Annotation, Location
from .client import DocumentCloud
from .documents import Document, DocumentClient, Mention
from .projects import Project, ProjectClient
from .sections import Section



"""
Documents
"""

# Future
from __future__ import division, print_function, unicode_literals

# Standard Library
import logging
import os
import re
import warnings
from functools import partial

# Third Party
from future.utils import python_2_unicode_compatible
from requests.exceptions import RequestException

# Local
from .annotations import AnnotationClient
from .base import APIResults, BaseAPIClient, BaseAPIObject
from .constants import BULK_LIMIT
from .exceptions import APIError
from .organizations import Organization
from .sections import SectionClient
from .toolbox import grouper, is_url, merge_dicts, requests_retry_session
from .users import User

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse



logger = logging.getLogger("documentcloud")


@python_2_unicode_compatible
class Document(BaseAPIObject):
    """A single DocumentCloud document"""

    api_path = "documents"
    writable_fields = [
        "access",
        "data",
        "description",
        "language",
        "published_url",
        "related_article",
        "source",
        "title",
    ]
    date_fields = ["created_at", "updated_at"]

    def __init__(self, client, dict_):

        # deal with potentially nested objects
        objs = [("user", User), ("organization", Organization)]
        for name, resource in objs:
            value = dict_.get(name)
            if isinstance(value, dict):
                dict_["_" + name] = resource(client, value)
                dict_[name + "_id"] = value.get("id")
            elif isinstance(value, int):
                dict_["_" + name] = None
                dict_[name + "_id"] = value

        super(Document, self).__init__(client, dict_)

        self.sections = SectionClient(client, self)
        self.annotations = AnnotationClient(client, self)
        self.notes = self.annotations

    def __str__(self):
        return self.title

    def __getattr__(self, attr):
        """Generate methods for fetching resources"""
        p_image = re.compile(
            r"^get_(?P<size>thumbnail|small|normal|large)_image_url(?P<list>_list)?$"
        )
        get = attr.startswith("get_")
        url = attr.endswith("_url")
        text = attr.endswith("_text")
        # this allows dropping `get_` to act like a property, ie
        # .full_text_url
        if not get and hasattr(self, "get_{}".format(attr)):
            return getattr(self, "get_{}".format(attr))()
        # this allows dropping `_url` to fetch the url, ie
        # .get_full_text()
        if not url and hasattr(self, "{}_url".format(attr)):
            return lambda *a, **k: self._get_url(
                getattr(self, "{}_url".format(attr))(*a, **k), text
            )
        # this genericizes the image sizes
        m_image = p_image.match(attr)
        if m_image and m_image.group("list"):
            return partial(self.get_image_url_list, size=m_image.group("size"))
        if m_image and not m_image.group("list"):
            return partial(self.get_image_url, size=m_image.group("size"))
        raise AttributeError(
            "'{}' object has no attribute '{}'".format(self.__class__.__name__, attr)
        )

    def __dir__(self):
        attrs = dir(type(self)) + list(self.__dict__.keys())
        getters = [a for a in attrs if a.startswith("get_")]
        attrs += [a[len("get_") :] for a in getters]
        attrs += [a[: -len("_url")] for a in getters if a.endswith("url")]
        attrs += [a[len("get_") : -len("_url")] for a in getters if a.endswith("url")]
        for size in ["thumbnail", "small", "normal", "large"]:
            attrs += [
                "get_{}_image_url".format(size),
                "{}_image_url".format(size),
                "get_{}_image".format(size),
                "{}_image".format(size),
                "get_{}_image_url_list".format(size),
                "{}_image_url_list".format(size),
            ]
        return sorted(attrs)

    @property
    def pages(self):
        return self.page_count

    @property
    def mentions(self):
        if hasattr(self, "highlights") and self.highlights is not None:
            return [
                Mention(page, text)
                for page, texts in self.highlights.items()
                for text in texts
            ]
        else:
            return []

    @property
    def user(self):
        # pylint:disable=access-member-before-definition
        if self._user is None:
            self._user = self._client.users.get(self.user_id)
        return self._user

    @property
    def organization(self):
        # pylint:disable=access-member-before-definition
        if self._organization is None:
            self._organization = self._client.organizations.get(self.organization_id)
        return self._organization

    @property
    def contributor(self):
        return self.user.name

    @property
    def contributor_organization(self):
        return self.organization.name

    @property
    def contributor_organization_slug(self):
        return self.organization.slug

    def _get_url(self, url, text):
        base_netloc = urlparse(self._client.base_uri).netloc
        url_netloc = urlparse(url).netloc

        if base_netloc == url_netloc:
            # if the url host is the same as the base api host,
            # sent the request with the client in order to include
            # authentication credentials
            response = self._client.get(url, full_url=True)
        else:
            response = requests_retry_session().get(
                url, headers={"User-Agent": "python-documentcloud2"}
            )
        if text:
            return response.text
        else:
            return response.content

    # Resource URLs
    def get_full_text_url(self):
        return "{}documents/{}/{}.txt".format(self.asset_url, self.id, self.slug)

    def get_page_text_url(self, page=1):
        return "{}documents/{}/pages/{}-p{}.txt".format(
            self.asset_url, self.id, self.slug, page
        )

    def get_json_text_url(self):
        return "{}documents/{}/{}.txt.json".format(self.asset_url, self.id, self.slug)

    def get_pdf_url(self):
        return "{}documents/{}/{}.pdf".format(self.asset_url, self.id, self.slug)

    def get_image_url(self, page=1, size="normal"):
        return "{}documents/{}/pages/{}-p{}-{}.gif".format(
            self.asset_url, self.id, self.slug, page, size
        )

    def get_image_url_list(self, size="normal"):
        return [
            self.get_image_url(page=i, size=size) for i in range(1, self.page_count + 1)
        ]

    def process(self):
        """Reprocess the document"""
        self._client.post("{}/{}/process/".format(self.api_path, self.id))


class DocumentClient(BaseAPIClient):
    """Client for interacting with Documents"""

    api_path = "documents"
    resource = Document

    def search(self, query, **params):
        """Return documents matching a search query"""

        mentions = params.pop("mentions", None)
        if mentions is not None:  # pragma: no cover
            warnings.warn(
                "The `mentions` argument to `search` is deprecated, "
                "it will always include mentions from all pages now",
                DeprecationWarning,
            )
        data = params.pop("data", None)
        if data is not None:  # pragma: no cover
            warnings.warn(
                "The `data` argument to `search` is deprecated, "
                "it will always include data now",
                DeprecationWarning,
            )

        if query:
            params["q"] = query
        response = self.client.get("documents/search/", params=params)
        return APIResults(self.resource, self.client, response)

    def upload(self, pdf, **kwargs):
        """Upload a document"""
        # if they pass in a URL, use the URL upload flow
        if is_url(pdf):
            return self._upload_url(pdf, **kwargs)
        # otherwise use the direct file upload flow - determine if they passed
        # in a file or a path
        elif hasattr(pdf, "read"):
            try:
                size = os.fstat(pdf.fileno()).st_size
            except (AttributeError, OSError):  # pragma: no cover
                size = 0
        else:
            size = os.path.getsize(pdf)
            pdf = open(pdf, "rb")

        # DocumentCloud's size limit is set to 501MB to give people a little leeway
        # for OS rounding
        if size >= 501 * 1024 * 1024:
            raise ValueError(
                "The pdf you have submitted is over the DocumentCloud API's 500MB "
                "file size limit. Split it into smaller pieces and try again."
            )

        return self._upload_file(pdf, **kwargs)

    def _format_upload_parameters(self, name, **kwargs):
        """Prepare upload parameters from kwargs"""
        allowed_parameters = [
            "access",
            "description",
            "language",
            "original_extension",
            "related_article",
            "published_url",
            "source",
            "title",
            "data",
            "force_ocr",
            "projects",
        ]
        # these parameters currently do not work, investigate...
        ignored_parameters = ["secure"]

        # title is required, so set a default
        params = {"title": self._get_title(name)}

        if "project" in kwargs:
            params["projects"] = [kwargs["project"]]

        for param in allowed_parameters:
            if param in kwargs:
                params[param] = kwargs[param]

        for param in ignored_parameters:
            if param in kwargs:
                warnings.warn(
                    "The parameter `{}` is not currently supported".format(param)
                )

        return params

    def _get_title(self, name):
        """Get the default title for a document from its path"""
        return name.split(os.sep)[-1].rsplit(".", 1)[0]

    def _upload_url(self, file_url, **kwargs):
        """Upload a document from a publicly accessible URL"""
        params = self._format_upload_parameters(file_url, **kwargs)
        params["file_url"] = file_url
        response = self.client.post("documents/", json=params)
        return Document(self.client, response.json())

    def _upload_file(self, file_, **kwargs):
        """Upload a document directly"""
        # create the document
        force_ocr = kwargs.pop("force_ocr", False)
        params = self._format_upload_parameters(file_.name, **kwargs)
        response = self.client.post("documents/", json=params)

        # upload the file directly to storage
        create_json = response.json()
        presigned_url = create_json["presigned_url"]
        response = requests_retry_session().put(presigned_url, data=file_.read())

        # begin processing the document
        doc_id = create_json["id"]
        response = self.client.post(
            "documents/{}/process/".format(doc_id), json={"force_ocr": force_ocr}
        )

        return Document(self.client, create_json)

    def _collect_files(self, path):
        """Find the paths to all pdfs under a directory"""
        path_list = []
        for (dirpath, _dirname, filenames) in os.walk(path):
            path_list.extend(
                [
                    os.path.join(dirpath, i)
                    for i in filenames
                    if i.lower().endswith(".pdf")
                ]
            )
        return path_list

    def upload_directory(self, path, handle_errors=False, **kwargs):
        """Upload all PDFs in a directory"""

        # do not set the same title for all documents
        kwargs.pop("title", None)

        # Loop through the path and get all the files
        path_list = self._collect_files(path)

        logger.info(
            "Upload directory on %s: Found %d files to upload", path, len(path_list)
        )

        # Upload all the pdfs using the bulk API to reduce the number
        # of API calls and improve performance
        obj_list = []
        params = self._format_upload_parameters("", **kwargs)
        for i, pdf_paths in enumerate(grouper(path_list, BULK_LIMIT)):
            # Grouper will put None's on the end of the last group
            pdf_paths = [p for p in pdf_paths if p is not None]

            logger.info("Uploading group %d: %s", i + 1, "\n".join(pdf_paths))

            # create the documents
            logger.info("Creating the documents...")
            try:
                response = self.client.post(
                    "documents/",
                    json=[
                        merge_dicts(params, {"title": self._get_title(p)})
                        for p in pdf_paths
                    ],
                )
            except (APIError, RequestException) as exc:
                if handle_errors:
                    logger.info(
                        "Error creating the following documents: %s %s",
                        exc,
                        "\n".join(pdf_paths),
                    )
                    continue
                else:
                    raise

            # upload the files directly to storage
            create_json = response.json()
            obj_list.extend(create_json)
            presigned_urls = [j["presigned_url"] for j in create_json]
            for url, pdf_path in zip(presigned_urls, pdf_paths):
                logger.info("Uploading %s to S3...", pdf_path)
                try:
                    response = requests_retry_session().put(
                        url, data=open(pdf_path, "rb").read()
                    )
                    self.client.raise_for_status(response)
                except (APIError, RequestException) as exc:
                    if handle_errors:
                        logger.info(
                            "Error uploading the following document: %s %s",
                            exc,
                            pdf_path,
                        )
                        continue
                    else:
                        raise

            # begin processing the documents
            logger.info("Processing the documents...")
            doc_ids = [j["id"] for j in create_json]
            try:
                response = self.client.post("documents/process/", json={"ids": doc_ids})
            except (APIError, RequestException) as exc:
                if handle_errors:
                    logger.info(
                        "Error creating the following documents: %s %s",
                        exc,
                        "\n".join(pdf_paths),
                    )
                    continue
                else:
                    raise

        logger.info("Upload directory complete")

        # Pass back the list of documents
        return [Document(self.client, d) for d in obj_list]


@python_2_unicode_compatible
class Mention:
    """A snippet from a document search"""

    def __init__(self, page, text):
        if page.startswith("page_no_"):
            page = page[len("page_no_") :]
        self.page = page
        self.text = text

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self)  # pragma: no cover

    def __str__(self):
        return '{} - "{}"'.format(self.page, self.text)



"""
The public interface for the DocumentCloud API
"""

# Future
from __future__ import division, print_function, unicode_literals

# Standard Library
import logging
from functools import partial

# Third Party
import ratelimit
import requests

# Local
from .constants import AUTH_URI, BASE_URI, RATE_LIMIT, RATE_PERIOD, TIMEOUT
from .documents import DocumentClient
from .exceptions import APIError, CredentialsFailedError, DoesNotExistError
from .organizations import OrganizationClient
from .projects import ProjectClient
from .toolbox import requests_retry_session
from .users import UserClient

logger = logging.getLogger("documentcloud")


class DocumentCloud(object):
    """
    The public interface for the DocumentCloud API
    """

    def __init__(
        self,
        username=None,
        password=None,
        base_uri=BASE_URI,
        auth_uri=AUTH_URI,
        timeout=TIMEOUT,
        loglevel=None,
        rate_limit=True,
    ):
        self.base_uri = base_uri
        self.auth_uri = auth_uri
        self.username = username
        self.password = password
        self._user_id = None
        self.timeout = timeout
        self.refresh_token = None
        self.session = requests.Session()
        self._set_tokens()

        if loglevel:  # pragma: no cover
            logging.basicConfig(
                level=loglevel,
                format="%(asctime)s %(levelname)-8s %(name)-25s %(message)s",
            )
        else:
            logger.addHandler(logging.NullHandler())

        self.documents = DocumentClient(self)
        self.projects = ProjectClient(self)
        self.users = UserClient(self)
        self.organizations = OrganizationClient(self)

        if rate_limit:
            self._request = ratelimit.limits(calls=RATE_LIMIT, period=RATE_PERIOD)(
                self._request
            )

    def _set_tokens(self):
        """Set the refresh and access tokens"""
        if self.refresh_token:
            access_token, self.refresh_token = self._refresh_tokens(self.refresh_token)
        elif self.username and self.password:
            access_token, self.refresh_token = self._get_tokens(
                self.username, self.password
            )
        else:
            access_token = None

        if access_token:
            self.session.headers.update(
                {"Authorization": "Bearer {}".format(access_token)}
            )

    def _get_tokens(self, username, password):
        """Get an access and refresh token in exchange for the username and password"""
        response = requests_retry_session().post(
            "{}token/".format(self.auth_uri),
            json={"username": username, "password": password},
            timeout=self.timeout,
        )

        if response.status_code == requests.codes.UNAUTHORIZED:
            raise CredentialsFailedError("The username and password is incorrect")

        self.raise_for_status(response)

        json = response.json()
        return (json["access"], json["refresh"])

    def _refresh_tokens(self, refresh_token):
        """Refresh the access and refresh tokens"""
        response = requests_retry_session().post(
            "{}refresh/".format(self.auth_uri),
            json={"refresh": refresh_token},
            timeout=self.timeout,
        )

        if response.status_code == requests.codes.UNAUTHORIZED:
            # refresh token is expired
            return self._get_tokens(self.username, self.password)

        self.raise_for_status(response)

        json = response.json()
        return (json["access"], json["refresh"])

    @property
    def user_id(self):
        if self._user_id is None:
            user = self.users.get("me")
            self._user_id = user.id
        return self._user_id

    def _request(self, method, url, raise_error=True, **kwargs):
        """Generic method to make API requests"""
        # pylint: disable=method-hidden
        logger.info("request: %s - %s - %s", method, url, kwargs)
        set_tokens = kwargs.pop("set_tokens", True)
        full_url = kwargs.pop("full_url", False)

        if not full_url:
            url = "{}{}".format(self.base_uri, url)

        response = requests_retry_session(session=self.session).request(
            method, url, timeout=self.timeout, **kwargs
        )
        logger.debug("response: %s - %s", response.status_code, response.content)
        if response.status_code == requests.codes.FORBIDDEN and set_tokens:
            self._set_tokens()
            # track set_tokens to not enter an infinite loop
            kwargs["set_tokens"] = False
            return self._request(method, url, full_url=True, **kwargs)

        if raise_error:
            self.raise_for_status(response)

        return response

    def __getattr__(self, attr):
        """Generate methods for each HTTP request type"""
        methods = ["get", "options", "head", "post", "put", "patch", "delete"]
        if attr in methods:
            return partial(self._request, attr)
        raise AttributeError(
            "'{}' object has no attribute '{}'".format(self.__class__.__name__, attr)
        )

    def raise_for_status(self, response):
        """Raise for status with a custom error class"""
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            if exc.response.status_code == 404:
                raise DoesNotExistError(response=exc.response)
            else:
                raise APIError(response=exc.response)



# Third Party
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='python-documentcloud',
    version='2.1.1',
    description='A simple Python wrapper for the DocumentCloud API',
    author='Mitchell Kotler',
    author_email='mitch@muckrock.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/muckrock/python-documentcloud',
    license="MIT",
    packages=("documentcloud",),
    include_package_data=True,
    install_requires=(
        'future',
        'listcrunch>=1.0.1',
        'python-dateutil',
        'ratelimit',
        'requests',
        'urllib3',
    ),
    extras_require={
        'dev': [
            'black',
            'coverage',
            'isort',
            'pylint',
            'sphinx',
            'twine',
        ],
        'test': [
            'pytest',
            'pytest-mock',
            'pytest-recording',
            'vcrpy',
        ],
    },
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
    )
)



"""
Custom exceptions for python-documentcloud
"""
# Future
from __future__ import division, print_function, unicode_literals


class DocumentCloudError(Exception):
    """Base class for errors for python-documentcloud"""

    def __init__(self, *args, **kwargs):
        self.response = kwargs.pop("response", None)
        if self.response is not None:
            self.error = self.response.text
            self.status_code = self.response.status_code
            if not args:
                args = ["{} - {}".format(self.status_code, self.error)]
        else:
            self.error = None
            self.status_code = None
        super(DocumentCloudError, self).__init__(*args, **kwargs)


class DuplicateObjectError(DocumentCloudError):
    """Raised when an object is added to a unique list more than once"""


class CredentialsFailedError(DocumentCloudError):
    """Raised if unable to obtain an access token due to bad login credentials"""


class APIError(DocumentCloudError):
    """Any other error calling the API"""


class DoesNotExistError(APIError):
    """Raised when the user asks the API for something it cannot find"""


class MultipleObjectsReturnedError(APIError):
    """Raised when the API returns multiple objects when it expected one"""
